/*
	Some maths functions

	Copyright (C) 2003	Patrice Mandin
	
	See COPYING in top directory for license
*/

/*--- Includes ---*/

#include <math.h>
#include <stdlib.h>

#include <SDL/SDL_opengl.h>

#include "bsp.h"
#include "calculations_types.h"
#include "camera.h"
#include "map_parse.h"
#include "map_draw.h"

/*--- Functions prototypes ---*/

static int CALC_DotProductPlus(point3_t *point, plane_t *plane);

/*--- Functions ---*/

int CALC_DotProduct(point3_t *point, plane_t *plane)
{
	return (
		(point->x*plane->x) +
		(point->y*plane->y) +
		(point->z*plane->z) -
		plane->dist
	);
}

static int CALC_DotProductPlus(point3_t *point, plane_t *plane)
{
	return (
		(point->x*plane->x) +
		(point->y*plane->y) +
		(point->z*plane->z) +
		plane->dist
	);
}

float CALC_GetLength(float x, float y, float z)
{
	return sqrt(x*x+y*y+z*z);
}

void CALC_NormalizePlane(plane_t *plane)
{
	float length;

	length = CALC_GetLength(plane->x, plane->y, plane->z);
	plane->x /= length;
	plane->y /= length;
	plane->z /= length;
	plane->dist /= length;
}

int CALC_BoundBoxViewable(bbox_t *bb)
{
	int i;

	/* Check each bounding box vertices against each clip plane */
	/*  No need to check against far plane ? */
	for (i=0; i<5; i++) {

		if (CALC_DotProductPlus(&(bb->vertexes[0]), &clip_planes[i])>0) {
			continue;
		}
		if (CALC_DotProductPlus(&(bb->vertexes[1]), &clip_planes[i])>0) {
			continue;
		}
		if (CALC_DotProductPlus(&(bb->vertexes[2]), &clip_planes[i])>0) {
			continue;
		}
		if (CALC_DotProductPlus(&(bb->vertexes[3]), &clip_planes[i])>0) {
			continue;
		}
		if (CALC_DotProductPlus(&(bb->vertexes[4]), &clip_planes[i])>0) {
			continue;
		}
		if (CALC_DotProductPlus(&(bb->vertexes[5]), &clip_planes[i])>0) {
			continue;
		}
		if (CALC_DotProductPlus(&(bb->vertexes[6]), &clip_planes[i])>0) {
			continue;
		}
		if (CALC_DotProductPlus(&(bb->vertexes[7]), &clip_planes[i])>=0) {
			continue;
		}

		return 0;
	}

	return 1;
}

int CALC_FaceViewable(bsp_face_t *cur_face, point3_t *origin)
{
	int i, j;
	int visible;
	point3_t cur_point;
	bsp_vertex_t	*cur_vertex;
	bsp_meshvertex_t	*cur_meshvertex;

	if (cur_face->type == FACE_BILLBOARD) {
		return 1;
	}

	/* Check each face vertices against each clip plane */
	/*  No need to check against far plane ? */
	for (i=0; i<5; i++) {

		visible=0;
		if ((cur_face->type == FACE_POLYGON) || (cur_face->type == FACE_PATCH)) {
			for (j=0; j<cur_face->n_vertices; j++) {
				cur_vertex=&(map.bsp.vertexes[cur_face->vertex + j]);
				cur_point=cur_vertex->position;
				if (origin!=NULL) {
					cur_point.x += origin->x;
					cur_point.y += origin->y;
					cur_point.z += origin->z;
				}
				if (CALC_DotProductPlus(&cur_point, &clip_planes[i])>0) {
					visible=1;
					break;
				}
			}
		} else if (cur_face->type == FACE_MESH) {
			cur_meshvertex = &(map.bsp.meshvertexes[cur_face->meshvert]);

			for (j=0; j<cur_face->n_meshverts; j++, cur_meshvertex++) {
				cur_vertex=&(map.bsp.vertexes[cur_face->vertex + cur_meshvertex->offset]);
				cur_point=cur_vertex->position;
				if (origin!=NULL) {
					cur_point.x += origin->x;
					cur_point.y += origin->y;
					cur_point.z += origin->z;
				}
				if (CALC_DotProductPlus(&cur_point, &clip_planes[i])>0) {
					visible=1;
					break;
				}
			}
		}

		if (!visible) {
			return 0;
		}
	}

	return 1;
}

int CALC_BoundBoxFaceOcculted(bbox_t *bb)
{
	int i, j, k;
	GLdouble	winx, winy, winz;
	bsp_face_t	*cur_face;
	bsp_vertex_t	*cur_v1, *cur_v2;
	int occulted;
	int v1, v2;
	float cur_side, side;
	GLdouble	ver1x, ver1y, ver1z;
	GLdouble	ver2x, ver2y, ver2z;

	/* Check if bounding box occulted entierely by one of the faces already drawn */

	for (i=0; i<map_render.num_faces_to_culltest; i++) {
		cur_face=&(map.bsp.faces[map_render.faces_to_draw[i]]);

		if (cur_face->type!=FACE_POLYGON) {
			continue;
		}

		/* Check the bounding box against this face */

		occulted=1;
		cur_side=side=0;
		for (j=0; j<8; j++) {
			if (gluProject(
				bb->vertexes[j].x, bb->vertexes[j].y, bb->vertexes[j].z,
				map_render.modelview, map_render.projection, map_render.viewport,
				&winx, &winy, &winz
					)==GL_FALSE) {
					/* If projection in viewport not possible */
					occulted=0;
					break;
			}

			/* Check if current vertex inside the face */
			v1 = cur_face->vertex;
			v2 = v1+1;
			for (k=0; k<cur_face->n_vertices; k++, v1++, v2++) {
				/* Last segment, check between last and first vertex */
				if (k==(cur_face->n_vertices)-1) {
					v2=0;
				}

				cur_v1=&(map.bsp.vertexes[v1]);
				cur_v2=&(map.bsp.vertexes[v2]);

				if (gluProject(
					cur_v1->position.x, cur_v1->position.y, cur_v1->position.z,
					map_render.modelview, map_render.projection, map_render.viewport,
					&ver1x, &ver1y, &ver1z
						)==GL_FALSE) {
						/* If projection in viewport not possible */
						occulted=0;
						break;
				}

				if (gluProject(
					cur_v2->position.x, cur_v2->position.y, cur_v2->position.z,
					map_render.modelview, map_render.projection, map_render.viewport,
					&ver2x, &ver2y, &ver2z
						)==GL_FALSE) {
						/* If projection in viewport not possible */
						occulted=0;
						break;
				}

				cur_side = (ver2x-ver1x)*(winy-ver1y) + (ver2y-ver1y)*(winx-ver1x); /* x1*y2 + x2*y1 */
				if (k==0) {
					side = cur_side;
				} else {
					/* This vertex not on the same face side ? */
					if (side != cur_side) {
						occulted=0;
						break;
					}
				}
			}
		}

		/* The 8 vertices of the bbox are occulted by the current face */
		if (occulted) {
			return 1;
		}
	}

	return 0;
}
