/*
	Map viewer

	Copyright (C) 2003	Patrice Mandin
	
	See COPYING in top directory for license
*/

/*--- Includes ---*/

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <SDL/SDL_endian.h>
#include <SDL/SDL_opengl.h>

#include "bsp.h"
#include "calculations_functions.h"
#include "camera.h"
#include "extensions.h"
#include "filesystem.h"
#include "map_parse.h"
#include "map_draw.h"
#include "parameters.h"
#include "statistics.h"
#include "verbose.h"
#include "viewer.h"

/*--- Defines ---*/

#define NUM_LIST_VERTICES	1
#define NUM_LIST_FACES	1
#define NUM_LIST_MODELS 1

#define MIPMAPPED_TEXTURES	1	/* To use mipmap for textures */

#define RENDERING_TEXTURED	(1<<0)
#define RENDERING_BLENDED	(1<<1)
#define RENDERING_TEXTURED2	(1<<2)
#define RENDERING_BLENDED2	(1<<3)
#define RENDERING_TEXARRAY	(0<<4)
#define RENDERING_LIGARRAY	(1<<4)

#define DISABLE_TEXTURING \
	if ((rendering_state & RENDERING_TEXTURED)==RENDERING_TEXTURED) { \
		rendering_state &= !RENDERING_TEXTURED; \
		glDisable(GL_TEXTURE_2D); \
	}

#define ENABLE_TEXTURING \
	if ((rendering_state & RENDERING_TEXTURED)!=RENDERING_TEXTURED) { \
		rendering_state |= RENDERING_TEXTURED; \
		glEnable(GL_TEXTURE_2D); \
	}

#define DISABLE_BLENDING \
	if ((rendering_state & RENDERING_BLENDED)==RENDERING_BLENDED) { \
		rendering_state &= !RENDERING_BLENDED; \
		glDisable(GL_BLEND); \
	}

#define ENABLE_BLENDING \
	if ((rendering_state & RENDERING_BLENDED)!=RENDERING_BLENDED) { \
		rendering_state |= RENDERING_BLENDED; \
		glEnable(GL_BLEND); \
	}

#define DISABLE_TEXTURING2 \
	if ((rendering_state & RENDERING_TEXTURED2)==RENDERING_TEXTURED2) { \
		glDisable(GL_TEXTURE_2D); \
		rendering_state &= !RENDERING_TEXTURED2; \
	}

#define ENABLE_TEXTURING2 \
	if ((rendering_state & RENDERING_TEXTURED2)!=RENDERING_TEXTURED2) { \
		glEnable(GL_TEXTURE_2D); \
		rendering_state |= RENDERING_TEXTURED2; \
	}

#define DISABLE_BLENDING2 \
	if ((rendering_state & RENDERING_BLENDED2)==RENDERING_BLENDED2) { \
		glDisable(GL_BLEND); \
		rendering_state &= !RENDERING_BLENDED2; \
	}

#define ENABLE_BLENDING2 \
	if ((rendering_state & RENDERING_BLENDED2)!=RENDERING_BLENDED2) { \
		glEnable(GL_BLEND); \
		rendering_state |= RENDERING_BLENDED2; \
	}

#define ENABLE_LIGHT_ARRAY \
	if ((rendering_state & RENDERING_LIGARRAY)!=RENDERING_LIGARRAY) { \
		glTexCoordPointer(2, GL_FLOAT, sizeof(bsp_vertex_t), &(map.bsp.vertexes->texcoord[1])); \
		rendering_state |= RENDERING_LIGARRAY; \
	}

#define ENABLE_TEX_ARRAY \
	if ((rendering_state & RENDERING_TEXARRAY)!=RENDERING_TEXARRAY) { \
		glTexCoordPointer(2, GL_FLOAT, sizeof(bsp_vertex_t), &(map.bsp.vertexes->texcoord[0])); \
		rendering_state &= !RENDERING_TEXARRAY; \
	}

/*--- Types ---*/

/*--- Global variables ---*/

map_render_t	map_render;

/*---- Variables ---*/

/* Textures */
static int *gl_textures=NULL;

/* Entities/models */
static entity_t *player_start;
static int list_models=-1;

/* Lightmaps */
static int *gl_lightmaps=NULL;

/* Faces */
static int list_faces=-1;
static unsigned long *face_to_patch;	/* Array to associate a face to a patch */

/* Patches */
static int max_patch_size;
static int list_patches=-1;

/* Misc */
static int camera_cluster, leaf_found;

/* Rendering flags */
static int rendering_state;

/*--- Functions prototypes ---*/

static int MAP_InitDrawTextures(void);
static int MAP_InitDrawFaces(void);
static int MAP_InitDrawModels(void);
static int MAP_InitDrawVertices(void);
static int MAP_InitDrawEntities(void);
static int MAP_InitDrawLightmaps(void);

static void MAP_ShutdownDrawTextures(void);
static void MAP_ShutdownDrawFaces(void);
static void MAP_ShutdownDrawModels(void);
static void MAP_ShutdownDrawLightmaps(void);
static void MAP_ShutdownDrawEntities(void);

static void MAP_RefreshDrawPatches(void);
static void MAP_PatchAddRaw(bsp_face_t	*cur_face, int sx, int sy);

static void MAP_DrawNode(int num_node);
static void MAP_DrawLeaf(int num_leaf);
static void MAP_DrawFaces(void);
static void MAP_DrawEntities(void);
static void MAP_DrawModels(bsp_leaf_t *leaf);

static int MAP_FaceIsTransparent(int n_texture);
static int MAP_SelectTexturing(int n_texture, int n_lightmap, int pass, int *last_rendering_pass);
static void MAP_DrawFace(int n_face, point3_t *origin);

/*--- Functions ---*/

void MAP_InitCamera(camera_t *newcamera, float *angle)
{
	if (player_start==NULL) {
		newcamera->from.x = map.minx + (map.maxx-map.minx)/2.0;
		newcamera->from.y = map.miny + (map.maxy-map.miny)/2.0;
		newcamera->from.z = map.minz + (map.maxz-map.minz)/2.0;
		*angle = 0.0;
	} else {
		newcamera->from.x = player_start->origin[0];
		newcamera->from.y = player_start->origin[1];
		newcamera->from.z = player_start->origin[2] + 64.0;
		*angle = player_start->angle;
	}
}

void MAP_InitDraw(void)
{
	MAP_InitDrawTextures();
	MAP_InitDrawLightmaps();
	MAP_InitDrawVertices();
	MAP_InitDrawFaces();
	MAP_InitDrawModels();
	MAP_InitDrawEntities();
}

static int MAP_InitDrawVertices(void)
{
	glVertexPointer(3, GL_FLOAT, sizeof(bsp_vertex_t), &(map.bsp.vertexes->position));
	glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(bsp_vertex_t), &(map.bsp.vertexes->color));
/*
	if (arb_multitexture_present && (multitexture_units>=2)) {
	    glClientActiveTextureARB(GL_TEXTURE1_ARB);
		glTexCoordPointer(2, GL_FLOAT, sizeof(bsp_vertex_t), &(map.bsp.vertexes->texcoord[1]));
	    glClientActiveTextureARB(GL_TEXTURE0_ARB);
	}
*/
	glTexCoordPointer(2, GL_FLOAT, sizeof(bsp_vertex_t), &(map.bsp.vertexes->texcoord[0]));
	rendering_state = RENDERING_TEXARRAY;
	glNormalPointer(GL_FLOAT, sizeof(bsp_vertex_t), &(map.bsp.vertexes->normal));

	VERBOSE(1, ("map: view: vertices ready\n"));
	return 1;
}

static int MAP_InitDrawFaces(void)
{
	bsp_face_t *cur_face;
	int i;

	map_render.face_flags=malloc(map.num_faces);
	if (map_render.face_flags==NULL) {
		fprintf(stderr,"map: can not allocate mem for face_flags\n");
		return 0;
	}

	map_render.faces_to_draw=malloc(sizeof(unsigned long)*map.num_faces);
	if (map_render.faces_to_draw==NULL) {
		fprintf(stderr,"map: can not allocate mem for faces_to_draw\n");
		return 0;
	}

	map_render.trans_to_draw=malloc(sizeof(unsigned long)*map.num_faces);
	if (map_render.trans_to_draw==NULL) {
		fprintf(stderr,"map: can not allocate mem for trans_to_draw\n");
		return 0;
	}

	list_faces = glGenLists(NUM_LIST_FACES);
	if (list_faces==0) {
		fprintf(stderr,"map: can not create gllist for faces\n");
		return 0;
	}

	/* Generate patches, and create a list for each one */
	/*  and calculate size of biggest patch */
	cur_face=map.bsp.faces;
	map.num_patches=0;
	max_patch_size=0;
	for (i=0; i<map.num_faces; i++, cur_face++) {
		if (cur_face->type==FACE_PATCH) {
			int size;

			++map.num_patches;

			size=cur_face->size[0] * cur_face->size[1];
			if (size>max_patch_size) max_patch_size = size;
		}
	}

	face_to_patch=malloc(sizeof(unsigned long)*map.num_faces);
	if (face_to_patch==NULL) {
		fprintf(stderr,"map: can not allocate mem for face_to_patch\n");
		return 0;
	}
	memset(face_to_patch, 0, sizeof(unsigned long)*map.num_faces);

	list_patches = glGenLists(map.num_patches);
	if (list_patches==0) {
		fprintf(stderr,"map: can not create gllist for patches\n");
		return 0;
	}

	MAP_RefreshDrawPatches();

	VERBOSE(1, ("map: view: %d faces ready\n", map.num_faces));
	VERBOSE(2, ("map: view:  %d patches ready\n", map.num_patches));
	return 1;
}

static int MAP_InitDrawTextures(void)
{
	int i;
	unsigned long memory_size;
	ftx_texture_t *cur_texture;

	gl_textures=(int *)malloc(sizeof(int)*map.num_textures);
	if (gl_textures==NULL) {
		fprintf(stderr,"map: can not allocate memory for textures\n");
		return 0;
	}

	glGenTextures(map.num_textures, gl_textures);

	memory_size=0;
	for (i=0; i<map.num_textures; i++) {
		cur_texture=(ftx_texture_t *)map.mem_textures[i];
		if (cur_texture==NULL) {
			continue;
		}	

		glBindTexture(GL_TEXTURE_2D, gl_textures[i]);

		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

		/*
		2d texture, level of detail 0 (normal), 3 components (red, green, blue),
		x size from image, y size from image, 
		border 0 (normal), rgb color data, unsigned byte data,
		and finally the data itself.
		*/

#if MIPMAPPED_TEXTURES
		gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, SDL_SwapLE32(cur_texture->width), SDL_SwapLE32(cur_texture->height), GL_RGBA, GL_UNSIGNED_BYTE, &(cur_texture->data));
#else
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, SDL_SwapLE32(cur_texture->width), SDL_SwapLE32(cur_texture->height), 0, GL_RGBA, GL_UNSIGNED_BYTE, &(cur_texture->data));
#endif

		memory_size += SDL_SwapLE32(cur_texture->width) * SDL_SwapLE32(cur_texture->height) * 4;
	}

	VERBOSE(1, ("map: view: textures ready\n"));
	VERBOSE(2, ("map: view:  %lu bytes used by textures\n", memory_size));
	return 1;
}

static int MAP_InitDrawLightmaps(void)
{
	int i;
	bsp_lightmap_t *cur_lightmap;
	unsigned long memory_size;

	gl_lightmaps=(int *)malloc(sizeof(int)*map.num_lightmaps);
	if (gl_lightmaps==NULL) {
		fprintf(stderr,"map: can not allocate memory for lightmaps\n");
		return 0;
	}

	glGenTextures(map.num_lightmaps, gl_lightmaps);

	memory_size=0;
	for (i=0; i<map.num_lightmaps; i++) {
		cur_lightmap=&(map.bsp.lightmaps[i]);

		glBindTexture(GL_TEXTURE_2D, gl_lightmaps[i]);

		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_NEAREST);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_REPEAT);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_REPEAT);

		/*
		2d texture, level of detail 0 (normal), 3 components (red, green, blue),
		x size from image, y size from image, 
		border 0 (normal), rgb color data, unsigned byte data,
		and finally the data itself.
		*/

#if MIPMAPPED_TEXTURES
		gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, 128, 128, GL_RGB, GL_UNSIGNED_BYTE, cur_lightmap);
#else
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 128, 128, 0, GL_RGB, GL_UNSIGNED_BYTE, cur_lightmap);
#endif

		memory_size += 128 * 128 * 4;
	}

	VERBOSE(1, ("map: view: lightmaps ready\n"));
	VERBOSE(2, ("map: view:  %lu bytes used by lightmaps\n", memory_size));
	return 1;
}

static int MAP_InitDrawEntities(void)
{
	entity_t *cur_entity;
	int i;

	list_models = glGenLists(NUM_LIST_MODELS);
	if (list_models==0) {
		fprintf(stderr,"map: can not create list\n");
		return 0;
	}

	cur_entity=map.mem_entities;
	player_start=NULL;
	for (i=0; i<map.num_entities; i++,cur_entity++) {
		if (strcmp("info_player_start",cur_entity->classname)==0) {
			player_start=cur_entity;
			VERBOSE(1, ("map: view: player start found\n"));
			break;
		}
	}

	VERBOSE(1, ("map: view: entities ready\n"));
	return 1;
}

static int MAP_InitDrawModels(void)
{
	map_render.model_flags=malloc(map.num_models);
	if (map_render.model_flags==NULL) {
		fprintf(stderr,"map: can not allocate mem for model_flags\n");
		return 0;
	}

	map_render.models_to_draw=malloc(sizeof(unsigned long)*map.num_models);
	if (map_render.models_to_draw==NULL) {
		fprintf(stderr,"map: can not allocate mem for models_to_draw\n");
		return 0;
	}

	VERBOSE(1, ("map: view: models ready\n"));
	return 1;
}

void MAP_ShutdownDraw(void)
{
	MAP_ShutdownDrawTextures();
	MAP_ShutdownDrawLightmaps();
	MAP_ShutdownDrawFaces();
	MAP_ShutdownDrawModels();
	MAP_ShutdownDrawEntities();
}

static void MAP_ShutdownDrawTextures(void)
{
	if (gl_textures!=NULL) {
		glDeleteTextures(map.num_textures, gl_textures);
		free(gl_textures);
		gl_textures=NULL;
	}
}

static void MAP_ShutdownDrawLightmaps(void)
{
	if (gl_lightmaps!=NULL) {
		glDeleteTextures(map.num_lightmaps, gl_lightmaps);
		free(gl_lightmaps);
		gl_lightmaps=NULL;
	}
}

static void MAP_ShutdownDrawFaces(void)
{
	if (map_render.face_flags!=NULL) {
		free(map_render.face_flags);
		map_render.face_flags=NULL;
	}

	if (map_render.trans_to_draw!=NULL) {
		free(map_render.trans_to_draw);
		map_render.trans_to_draw=NULL;
	}

	if (map_render.faces_to_draw!=NULL) {
		free(map_render.faces_to_draw);
		map_render.faces_to_draw=NULL;
	}

	if (face_to_patch!=NULL) {
		free(face_to_patch);
		face_to_patch=NULL;
	}

	if (list_patches!=-1) {
		glDeleteLists( list_patches, map.num_patches);
		list_patches=-1;
	}

	if (list_faces!=-1) {
		glDeleteLists( list_faces, NUM_LIST_FACES);
		list_faces=-1;
	}
}

static void MAP_ShutdownDrawModels(void)
{
	if (map_render.model_flags!=NULL) {
		free(map_render.model_flags);
		map_render.model_flags=NULL;
	}

	if (map_render.models_to_draw!=NULL) {
		free(map_render.models_to_draw);
		map_render.models_to_draw=NULL;
	}
}

static void MAP_ShutdownDrawEntities(void)
{
	player_start=NULL;

	if (list_models!=-1) {
		glDeleteLists( list_models, NUM_LIST_MODELS);
		list_models=-1;
	}
}

void MAP_RefreshDraw(void)
{
	MAP_RefreshDrawPatches();
}

static void MAP_RefreshDrawPatches(void)
{
	bsp_face_t	*cur_face;
	int i,j;
	int sx,sy;
	int x,y;
	int px,py;
	float	*vtx_control, *cur_vtx_ctl;
	float	*tex_control, *cur_tex_ctl;
	float	*nor_control, *cur_nor_ctl;
	float	*col_control, *cur_col_ctl;

	vtx_control=(float *)malloc(sizeof(float)*max_patch_size*(3+2+3+4));
	if (vtx_control==NULL) {
		fprintf(stderr,"map: view: can not allocate memory for patch generator\n");
		return;
	}
	tex_control=vtx_control+max_patch_size*3;
	nor_control=vtx_control+max_patch_size*(3+2);
	col_control=vtx_control+max_patch_size*(3+2+3);

	cur_face=map.bsp.faces;
	j=0;
	for (i=0; i<map.num_faces; i++, cur_face++) {
		if (cur_face->type!=FACE_PATCH) {
			continue;
		}

		face_to_patch[i]=j;

		sx=cur_face->size[0];
		sy=cur_face->size[1];

		/* Copy vertices in control array */
		cur_vtx_ctl=vtx_control;
		cur_tex_ctl=tex_control;
		cur_nor_ctl=nor_control;
		cur_col_ctl=col_control;
		for (y=0; y<sy; y++) {
			for (x=0; x<sx; x++) {
				int r,g,b,a;

				px=x /*(sx-1)-x*/;
				py=y;

				*cur_vtx_ctl++=map.bsp.vertexes[cur_face->vertex + px + py*sx].position.x;
				*cur_vtx_ctl++=map.bsp.vertexes[cur_face->vertex + px + py*sx].position.y;
				*cur_vtx_ctl++=map.bsp.vertexes[cur_face->vertex + px + py*sx].position.z;

				switch(polyfill_type) {
					case POLYFILL_LIGHTMAP:
						*cur_tex_ctl++=map.bsp.vertexes[cur_face->vertex + px + py*sx].texcoord[1].x;
						*cur_tex_ctl++=map.bsp.vertexes[cur_face->vertex + px + py*sx].texcoord[1].y;
						break;
					default:
						*cur_tex_ctl++=map.bsp.vertexes[cur_face->vertex + px + py*sx].texcoord[0].x;
						*cur_tex_ctl++=map.bsp.vertexes[cur_face->vertex + px + py*sx].texcoord[0].y;
						break;
				}

				*cur_nor_ctl++=map.bsp.vertexes[cur_face->vertex + px + py*sx].normal.x;
				*cur_nor_ctl++=map.bsp.vertexes[cur_face->vertex + px + py*sx].normal.y;
				*cur_nor_ctl++=map.bsp.vertexes[cur_face->vertex + px + py*sx].normal.z;

				r=(map.bsp.vertexes[cur_face->vertex + px + py*sx].color>>24)&255;
				g=(map.bsp.vertexes[cur_face->vertex + px + py*sx].color>>16)&255;
				b=(map.bsp.vertexes[cur_face->vertex + px + py*sx].color>>8)&255;
				a=(map.bsp.vertexes[cur_face->vertex + px + py*sx].color)&255;
				*cur_col_ctl++=(float)a/255;
				*cur_col_ctl++=(float)b/255;
				*cur_col_ctl++=(float)g/255;
				*cur_col_ctl++=(float)r/255;
			}
		}

		glNewList(list_patches+j, GL_COMPILE);
		glEnable(GL_MAP2_VERTEX_3);
		glEnable(GL_MAP2_TEXTURE_COORD_2);
		glEnable(GL_MAP2_NORMAL);
		glEnable(GL_MAP2_COLOR_4);

		for (y=0; y<(sy-1)>>1; y++) {
			py=y*2;
			for (x=0; x<(sx-1)>>1; x++) {
				px=x*2;

				glMap2f(
					GL_MAP2_VERTEX_3,
					0.0, 1.0, 3, 3,
					0.0, 1.0, sx*3, 3,
					&vtx_control[(py*sx+px)*3]
				);
				glMap2f(
					GL_MAP2_TEXTURE_COORD_2,
					0.0, 1.0, 2, 2,
					0.0, 1.0, sx*2, 2,
					&tex_control[(py*sx+px)*2]
				);
				glMap2f(
					GL_MAP2_NORMAL,
					0.0, 1.0, 3, 3,
					0.0, 1.0, sx*3, 3,
					&nor_control[(py*sx+px)*3]
				);
				glMap2f(
					GL_MAP2_COLOR_4,
					0.0, 1.0, 4, 4,
					0.0, 1.0, sx*4, 4,
					&col_control[(py*sx+px)*4]
				);

				glMapGrid2f(
					1+(1<<(tesselation+1)), 0.0, 1.0,
					1+(1<<(tesselation+1)), 0.0, 1.0
				);
				glEvalMesh2(
					GL_FILL,
					0, 1+(1<<(tesselation+1)),
					0, 1+(1<<(tesselation+1))
				);
			}
		}
/*		MAP_PatchAddRaw(cur_face, sx, sy);*/
		glDisable(GL_MAP2_VERTEX_3);
		glDisable(GL_MAP2_TEXTURE_COORD_2);
		glDisable(GL_MAP2_NORMAL);
		glDisable(GL_MAP2_COLOR_4);
		glEndList();

		j++;
	}

	free(vtx_control);
}

static void MAP_PatchAddRaw(bsp_face_t	*cur_face, int sx, int sy)
{
	int x,y;
/*
	DISABLE_TEXTURING;
	DISABLE_BLENDING;
	glPolygonMode(GL_FRONT,GL_LINE);
*/
	for (y=0; y<sy-1; y++) {
		for (x=0; x<sx-1; x++) {
			glBegin(GL_TRIANGLE_FAN);
			glArrayElement(cur_face->vertex + x + y*sx);
			glArrayElement(cur_face->vertex + x + (y+1)*sx);
			glArrayElement(cur_face->vertex + (x+1) + (y+1)*sx);
			glArrayElement(cur_face->vertex + (x+1) + y*sx);
			glEnd();
		}
	}
/*
	glPolygonMode(GL_FRONT,GL_FILL);
*/
}

void MAP_Draw(void)
{
	STAT_Clear();

	rendering_state=RENDERING_TEXARRAY;
	if ((arb_multitexture_present) && (multitexture_units>=2)) {
		glActiveTextureARB(GL_TEXTURE1_ARB);
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_BLEND);
	    glClientActiveTextureARB(GL_TEXTURE1_ARB);
		glTexCoordPointer(2, GL_FLOAT, sizeof(bsp_vertex_t), &(map.bsp.vertexes->texcoord[1]));
		glActiveTextureARB(GL_TEXTURE0_ARB);
	    glClientActiveTextureARB(GL_TEXTURE0_ARB);
	}
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glTexCoordPointer(2, GL_FLOAT, sizeof(bsp_vertex_t), &(map.bsp.vertexes->texcoord[0]));

	glGetDoublev(GL_MODELVIEW_MATRIX, map_render.modelview);
	glGetDoublev(GL_PROJECTION_MATRIX, map_render.projection);
	glGetIntegerv(GL_VIEWPORT, map_render.viewport);

	memset(map_render.face_flags,0, map.num_faces);
	map_render.num_faces_to_draw=0;
	map_render.num_trans_to_draw=0;
	map_render.num_faces_to_culltest=0;
	memset(map_render.model_flags, MODEL_UNCHECKED, map.num_models);
	map_render.num_models_to_draw=0;

	leaf_found=0;
	MAP_DrawNode(0);
	MAP_DrawFaces();
	glCallList(list_faces);

	if (models_enabled) {
		int i;

		MAP_DrawEntities();
		glCallList(list_models);

		stats.models.entry=map.num_models;
		for (i=0; i<map.num_models; i++) {
			if (map_render.model_flags[i]==MODEL_FRUSTUM_CULLED) {
				++stats.models.bbox;
			}
		}
	}
}

static void MAP_DrawNode(int num_node)
{
	bsp_node_t *cur_node;
	bsp_plane_t *cur_plane;
	int cur_child;

	++stats.nodes.entry;

	cur_node=&(map.bsp.nodes[num_node]);

	/* Check if node in view frustum */
	if (!CALC_BoundBoxViewable(&(map.bbox_nodes[num_node]))) {
		return;
	}

	++stats.nodes.bbox;

	/* Check if node occulted by already drawn faces */
/*
	if (CALC_BoundBoxFaceOcculted(&(map.bbox_nodes[num_node]))==1) {
		return;
	}
*/
	++stats.nodes.faces;

	/* Check if camera is front/back of node split */
	cur_plane=&(map.bsp.planes[cur_node->plane]);
	cur_child=(CALC_DotProduct(&(camera.from),&(cur_plane->plane)) < 0);

	/* Draw front leaf */
	if (cur_node->children[cur_child]>=0) {
		MAP_DrawNode(cur_node->children[cur_child]);
	} else {
		MAP_DrawLeaf(-(cur_node->children[cur_child]+1));
	}

	/* Draw back leaf */
	cur_child ^=1;
	if (cur_node->children[cur_child]>=0) {
		MAP_DrawNode(cur_node->children[cur_child]);
	} else {
		MAP_DrawLeaf(-(cur_node->children[cur_child]+1));
	}
}

static void MAP_DrawLeaf(int num_leaf)
{
	bsp_leaf_t	*cur_leaf;
	bsp_leafface_t *cur_leafface;
	unsigned char *cur_visdata;
	long cur_cluster;
	int i, cur_nface;

	++stats.leaves.entry;

	cur_leaf = &(map.bsp.leaves[num_leaf]);

	cur_cluster=cur_leaf->cluster;

	/* Memorize camera leaf, to check for visdata */
	if (!leaf_found) {
		camera_cluster=cur_cluster;
		leaf_found=1;
	}

	/* Check camera leaf against this one */	
	if (cur_cluster<0) {
		return;
	}

	cur_visdata=&(map.bsp.visdata->visdata);
	if ((cur_visdata[(camera_cluster*map.bsp.visdata->sz_vecs)+(cur_cluster>>3)] & (1<<(cur_cluster & 7)))==0) {
		return;
	}

	++stats.leaves.pvs;

	/* Check if leaf in view frustum */
	if (!CALC_BoundBoxViewable(&(map.bbox_leaves[num_leaf]))) {
		return;
	}

	++stats.leaves.bbox;

	/* Add faces to list draw */
	cur_leafface = &(map.bsp.leaffaces[cur_leaf->leafface]);
	for (i=0; i<cur_leaf->n_leaffaces; i++, cur_leafface++) {
		cur_nface=cur_leafface->face;
		if (map_render.face_flags[cur_nface]==0) {
			if (MAP_FaceIsTransparent(map.bsp.faces[cur_nface].texture)) {
				map_render.trans_to_draw[map_render.num_trans_to_draw++]=cur_nface;
			} else {
				map_render.faces_to_draw[map_render.num_faces_to_draw++]=cur_nface;
			}
			map_render.face_flags[cur_nface]=1;
/*
			if (map_render.num_faces_to_culltest==0) {
				map_render.num_faces_to_culltest=map_render.num_faces_to_draw;
			}
*/
		}
	}

	/* Add models to list draw */
	MAP_DrawModels(cur_leaf);
}

static void MAP_DrawFaces(void)
{
	int i;

	glNewList( list_faces, GL_COMPILE);

	stats.render.opaque=map_render.num_faces_to_draw;
	stats.render.transparent=map_render.num_trans_to_draw;

	/* First opaque faces, front to back */
	for (i=0; i<map_render.num_faces_to_draw; i++) {
		MAP_DrawFace(map_render.faces_to_draw[i], NULL);
	}

	/* Then draw transparent faces, back to front */
	glDepthMask(GL_FALSE);
	for (i=0; i<map_render.num_trans_to_draw; i++) {
		MAP_DrawFace(map_render.trans_to_draw[map_render.num_trans_to_draw-1-i], NULL);
	}
	glDepthMask(GL_TRUE);

	glEndList();
}

static void MAP_DrawModels(bsp_leaf_t *leaf)
{
	entity_t *cur_entity;
	bsp_model_t *cur_model;
	int i, j, num_model;
	int isinbox;
	bbox_t	cur_bbox;

	/* Already want to draw all models ? */
/*
	if (map_render.num_models_to_draw>=map.num_models-1) {
		return;
	}
*/
	cur_model = map.bsp.models;
	for (i=1; i<map.num_models; i++, cur_model++) {
		num_model=i;
		cur_entity=map.model_to_entity[num_model];

		if (map_render.model_flags[num_model]!=MODEL_UNCHECKED) {
			continue;
		}

		/* Check if model in view frustum */
		for (j=0; j<8; j++) {
			cur_bbox.vertexes[j].x = map.bbox_models[num_model].vertexes[j].x + cur_entity->origin[0];
			cur_bbox.vertexes[j].y = map.bbox_models[num_model].vertexes[j].y + cur_entity->origin[1];
			cur_bbox.vertexes[j].z = map.bbox_models[num_model].vertexes[j].z + cur_entity->origin[2];
		}

		if (CALC_BoundBoxViewable(&cur_bbox)==0) {
			map_render.model_flags[num_model]=MODEL_FRUSTUM_CULLED;	/* model won't be viewable at all */
			continue;
		}

		/* Check if model is in part of the leaf */
		isinbox=0;
		for (j=0; j<8; j++) {
			point3_t	cur_vertex;

			cur_vertex.x = map.bbox_models[num_model].vertexes[j].x + cur_entity->origin[0];
			cur_vertex.y = map.bbox_models[num_model].vertexes[j].y + cur_entity->origin[1];
			cur_vertex.z = map.bbox_models[num_model].vertexes[j].z + cur_entity->origin[2];

			/* Check if this model vertex in the leaf */
			if (
				(cur_vertex.x>=leaf->mins[0]) && (cur_vertex.x<=leaf->maxs[0]) &&
				(cur_vertex.y>=leaf->mins[1]) && (cur_vertex.y<=leaf->maxs[1]) &&
				(cur_vertex.z>=leaf->mins[2]) && (cur_vertex.z<=leaf->maxs[2])) {
				isinbox=1;
				break;
			}
		}

		/* Add model to the list of models to draw ? */
		if (!isinbox) {
			continue;
		}
		
		++stats.models.leaf;

		map_render.models_to_draw[map_render.num_models_to_draw++]=num_model;
		map_render.model_flags[num_model]=MODEL_VISIBLE_LEAF;	/* model visible from this leaf */
	}
}

static void MAP_DrawEntities(void)
{
	entity_t *cur_entity;
	bsp_model_t *cur_model;
	point3_t	cur_origin;
	int i,j, num_model;

	glNewList( list_models, GL_COMPILE);
	glMatrixMode(GL_MODELVIEW);

	for (i=0; i<map_render.num_models_to_draw; i++) {
		num_model=map_render.models_to_draw[i];

		cur_model=&(map.bsp.models[num_model]);
		cur_entity=map.model_to_entity[num_model];

		/* Draw only wanted model */
		if (model_selected!=-1) {
			if (num_model!=model_selected) {
				continue;
			}
		}

		cur_origin.x = cur_entity->origin[0];
		cur_origin.y = cur_entity->origin[1];
		cur_origin.z = cur_entity->origin[2];

		glPushMatrix();
		glTranslatef(cur_origin.x, cur_origin.y, cur_origin.z);

		for (j=0; j<cur_model->n_faces; j++) {
			MAP_DrawFace(cur_model->face + j, &cur_origin);
		}

		glPopMatrix();
	}

	glEndList();
}

static int MAP_FaceIsTransparent(int n_texture)
{
	/* No texture ? */
	if ((n_texture<0) || (map.mem_textures[n_texture]==NULL)) {
		return 1;
	}

	/* Transparent flag ? */
	if ((map.bsp.textures[n_texture].contents & 0x20000000)!=0) {
		return 1;
	}

	return 0;
}

static int MAP_SelectTexturing(int n_texture, int n_lightmap, int pass, int *last_rendering_pass)
{
	int cur_polyfill_type;

	*last_rendering_pass=SDL_TRUE;

	/* No texture ? */
	if ((n_texture<0) || (map.mem_textures[n_texture]==NULL)) {
		return 0;
	}

	/* No lightmap ? */
	cur_polyfill_type = polyfill_type;
	if (n_lightmap<0) {
		switch(cur_polyfill_type) {
			case POLYFILL_LIGHTMAP:
				cur_polyfill_type=POLYFILL_TEXTURE;
				break;
			case POLYFILL_TEXMODLIGHT:
				cur_polyfill_type=POLYFILL_TEXMODVER;
				break;
		}
	}

	/* Select texturing */
	switch(cur_polyfill_type) {
		case POLYFILL_VERTEXCOLOR:
			if (arb_multitexture_present && (multitexture_units>=2)) {
				glActiveTextureARB(GL_TEXTURE1_ARB);
				DISABLE_TEXTURING2;
				DISABLE_BLENDING2;
				glActiveTextureARB(GL_TEXTURE0_ARB);
			}
			DISABLE_TEXTURING;
			DISABLE_BLENDING;
			break;
		case POLYFILL_TEXTURE:
			ENABLE_TEXTURING;
			if ((map.bsp.textures[n_texture].contents & 0x20000000)==0) {
				DISABLE_BLENDING;
			} else {
				ENABLE_BLENDING;
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			}
			glBindTexture(GL_TEXTURE_2D, gl_textures[n_texture]);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
			ENABLE_TEX_ARRAY;
			break;
		case POLYFILL_TEXMODVER:
			ENABLE_TEXTURING;
			if ((map.bsp.textures[n_texture].contents & 0x20000000)==0) {
				DISABLE_BLENDING;
			} else {
				ENABLE_BLENDING;
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			}
			glBindTexture(GL_TEXTURE_2D, gl_textures[n_texture]);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			ENABLE_TEX_ARRAY;
			break;
		case POLYFILL_LIGHTMAP:
			ENABLE_TEXTURING;
			if ((map.bsp.textures[n_texture].contents & 0x20000000)==0) {
				DISABLE_BLENDING;
			} else {
				ENABLE_BLENDING;
				glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
			}
			glBindTexture(GL_TEXTURE_2D, gl_lightmaps[n_lightmap]);
			glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
			ENABLE_LIGHT_ARRAY;
			break;
		case POLYFILL_TEXMODLIGHT:
			if (arb_multitexture_present && (multitexture_units>=2)) {
				glActiveTextureARB(GL_TEXTURE1_ARB);
				ENABLE_TEXTURING2;
				if ((map.bsp.textures[n_texture].contents & 0x20000000)==0) {
					DISABLE_BLENDING2;
				} else {
					ENABLE_BLENDING2;
					glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				}
				glBindTexture(GL_TEXTURE_2D, gl_lightmaps[n_lightmap]);
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
			    glClientActiveTextureARB(GL_TEXTURE1_ARB);
				ENABLE_LIGHT_ARRAY;

				glActiveTextureARB(GL_TEXTURE0_ARB);
				ENABLE_TEXTURING;
				if ((map.bsp.textures[n_texture].contents & 0x20000000)==0) {
					DISABLE_BLENDING;
				} else {
					ENABLE_BLENDING;
					glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
				}
				glBindTexture(GL_TEXTURE_2D, gl_textures[n_texture]);
				glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
			    glClientActiveTextureARB(GL_TEXTURE0_ARB);
				ENABLE_TEX_ARRAY;
			} else {
				switch(pass) {
					case 0:
						ENABLE_TEXTURING;
						DISABLE_BLENDING;
						glBindTexture(GL_TEXTURE_2D, gl_textures[n_texture]);
						glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
						ENABLE_TEX_ARRAY;
						*last_rendering_pass=SDL_FALSE;
						break;
					case 1:
						ENABLE_TEXTURING;
						ENABLE_BLENDING;
						glBindTexture(GL_TEXTURE_2D, gl_lightmaps[n_lightmap]);
						glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
						ENABLE_LIGHT_ARRAY;
						glBlendFunc(GL_ZERO, GL_SRC_COLOR);
						break;
				}
			}
			break;
	}

	return 1;
}

static void MAP_DrawFace(int n_face, point3_t *origin)
{
	bsp_face_t *cur_face;
	int n_pass, last_pass;

	cur_face=&(map.bsp.faces[n_face]);

	n_pass=0;
	do {
		if (!MAP_SelectTexturing(cur_face->texture, cur_face->lm_index, n_pass, &last_pass)) {
			break;
		}

		++n_pass;

		switch(cur_face->type) {
			case FACE_POLYGON:
				if (last_pass) {
					++stats.faces.polys_entry;
				}
				if (CALC_FaceViewable(cur_face, origin)>0) {
					glDrawArrays(GL_TRIANGLE_FAN, cur_face->vertex, cur_face->n_vertices);
					if (last_pass) {
						++stats.faces.polys_bbox;
					}
				}
				break;
			case FACE_MESH:
				{
					int i;
					bsp_meshvertex_t *cur_meshvertex;

					if (last_pass) {
						++stats.faces.meshes_entry;
					}

					if (CALC_FaceViewable(cur_face, origin)>0) {
						cur_meshvertex = &(map.bsp.meshvertexes[cur_face->meshvert]);

						glBegin(GL_TRIANGLES);
						for (i=0; i<cur_face->n_meshverts; i++, cur_meshvertex++) {
							glArrayElement(cur_face->vertex + cur_meshvertex->offset);
						}
						glEnd();

						if (last_pass) {
							++stats.faces.meshes_bbox;
						}
					}
				}
				break;
			case FACE_PATCH:
				{
					if (last_pass) {
						++stats.faces.patches_entry;
					}
					if (CALC_FaceViewable(cur_face, origin)>0) {
						if (patch_selected==-1) {
							glCallList(list_patches+face_to_patch[n_face]);
						} else if (patch_selected==face_to_patch[n_face]) {
							glCallList(list_patches+face_to_patch[n_face]);
						}
					}
					if (last_pass) {
						++stats.faces.patches_bbox;
					}
				}
				break;
/*
			case FACE_BILLBOARD:
				if (last_pass) ++stats.faces.billboards;
				break;
*/
			default:
				break;
		}
	} while (last_pass!=SDL_TRUE);
}
