/*
	Map loader/parser

	Copyright (C) 2003	Patrice Mandin
	
	See COPYING in top directory for license
*/

/*--- Includes ---*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <physfs.h>
#include <SDL/SDL.h>
#include <SDL/SDL_endian.h>

#include "bsp.h"
#include "calculations_functions.h"
#include "filesystem.h"
#include "map_draw.h"
#include "map_parse.h"
#include "parameters.h"
#include "verbose.h"

/*--- Defines ---*/

/*--- Types ---*/

/*--- Global variables ---*/

map_t	map;

/*--- Functions prototypes ---*/

static void *MAP_GetPointer(int fakk_entry);
static int MAP_Endianize(void);

static int MAP_ParseEntities(void);
static int MAP_ParseTextures(void);
static int MAP_ParseVertices(void);
static int MAP_ParseFaces(void);
static int MAP_ParseBoundingBoxes(void);
static int MAP_ParseLightmaps(void);
static int MAP_ParseModels(void);	/* Must be done after ParseEntities */

static int MAP_ParseDummies(void);

/*--- Functions ---*/

int MAP_Load(char *filename)
{
	char *curfilename;

	map.bsp.header=NULL;
	
	map.bsp.textures=NULL;
	map.bsp.planes=NULL;
	map.bsp.lightmaps=NULL;
	map.bsp.faces=NULL;
	map.bsp.vertexes=NULL;
	map.bsp.meshvertexes=NULL;
	map.bsp.leafbrushes=NULL;
	map.bsp.leaffaces=NULL;
	map.bsp.leaves=NULL;
	map.bsp.nodes=NULL;
	map.bsp.brushsides=NULL;
	map.bsp.brushes=NULL;
	map.bsp.effects=NULL;
	map.bsp.models=NULL;
	map.bsp.entities=NULL;
	map.bsp.visdata=NULL;
	map.bsp.lightvols=NULL;

	map.bsp_bbox=map.bbox_nodes=map.bbox_leaves=map.bbox_models=NULL;

	map.num_textures=-1;
	map.mem_textures=NULL;

	map.num_entities=-1;
	map.mem_entities=NULL;

	map.num_faces=-1;

	map.num_lightmaps=-1;

	map.num_models=-1;
	map.model_to_entity=NULL;

	curfilename=malloc(strlen(filename)+1+5+4);
	if (curfilename==NULL) {
		fprintf(stderr,"map: not enough memory for filename");
		return 0;
	}

	sprintf(curfilename, "maps/%s.bsp", filename);
	map.bsp.header = (bsp_header_t *)FS_Load(curfilename, NULL);

	free(curfilename);
	return 1;
}

int MAP_Unload(void)
{
	MAP_ShutdownDraw();

	map.num_faces=-1;

	map.num_lightmaps=-1;

	if (map.mem_textures!=NULL) {
		int i;

		for (i=0; i<map.num_textures; i++) {
			if (map.mem_textures[i] != NULL) {
				free(map.mem_textures[i]);
				map.mem_textures[i]=NULL;
			}
		}

		map.num_textures=-1;
		free(map.mem_textures);
		map.mem_textures=NULL;
	}
	
	if (map.mem_entities!=NULL) {
		free(map.mem_entities);
		map.mem_entities=NULL;
		map.num_entities=-1;
	}

	if (map.model_to_entity!=NULL) {
		free(map.model_to_entity);
		map.model_to_entity=NULL;
		map.num_models=-1;
	}

	if (map.bsp_bbox!=NULL) {
		free(map.bsp_bbox);
		map.bsp_bbox=NULL;
	}

	if (map.bsp.header!=NULL) {
		free(map.bsp.header);
		map.bsp.header=NULL;
	}

	return 1;
}

static void *MAP_GetPointer(int fakk_entry)
{
	if (map.bsp.header==NULL) {
		fprintf(stderr, "map: access to NULL map_bsp\n");
		return NULL;
	}

	return (((unsigned char *)map.bsp.header) + map.bsp.header->entries[fakk_entry].offset);
}

int MAP_Parse(void)
{
	if (map.bsp.header==NULL) {
		fprintf(stderr,"map: No bsp to parse\n");
		return 0;
	}

	MAP_Endianize();

	/* Check for FAKK */
	if (map.bsp.header->ident!=IDENT_FAKK) {
		fprintf(stderr,"map: Unknown bsp ID: 0x%08lx\n",map.bsp.header->ident);
		return 0;
	}

	/* Check for version */
	if (map.bsp.header->version!=IDENT_VERSION) {
		fprintf(stderr,"map: Unknown bsp version: 0x%08lx\n",map.bsp.header->version);
		return 0;
	}

	map.bsp.textures=(bsp_texture_t *)MAP_GetPointer(FAKK_TEXTURES);
	map.bsp.planes=(bsp_plane_t *)MAP_GetPointer(FAKK_PLANES);
	map.bsp.lightmaps=(bsp_lightmap_t *)MAP_GetPointer(FAKK_LIGHTMAPS);
	map.bsp.faces=(bsp_face_t *)MAP_GetPointer(FAKK_FACES);
	map.bsp.vertexes = (bsp_vertex_t *) MAP_GetPointer(FAKK_VERTEXES);
	map.bsp.meshvertexes = (bsp_meshvertex_t *) MAP_GetPointer(FAKK_MESHVERTEXES);
	map.bsp.leafbrushes = (bsp_leafbrush_t *) MAP_GetPointer(FAKK_LEAFBRUSHES);
	map.bsp.leaffaces = (bsp_leafface_t *) MAP_GetPointer(FAKK_LEAFFACES);
	map.bsp.leaves=(bsp_leaf_t *)MAP_GetPointer(FAKK_LEAVES);
	map.bsp.nodes=(bsp_node_t *)MAP_GetPointer(FAKK_NODES);
	map.bsp.brushsides=(bsp_brushside_t *)MAP_GetPointer(FAKK_BRUSHSIDES);
	map.bsp.brushes=(bsp_brush_t *)MAP_GetPointer(FAKK_BRUSHES);
	map.bsp.effects=(bsp_effect_t *)MAP_GetPointer(FAKK_EFFECTS);
	map.bsp.models = (bsp_model_t *) MAP_GetPointer(FAKK_MODELS);
	map.bsp.entities = (unsigned char *) MAP_GetPointer(FAKK_ENTITIES);
	map.bsp.visdata=(bsp_visdata_t *)MAP_GetPointer(FAKK_VISDATA);
	map.bsp.lightvols=(bsp_lightvol_t *)MAP_GetPointer(FAKK_LIGHTVOLS);

	MAP_ParseEntities();
	MAP_ParseTextures();
	MAP_ParseVertices();
	MAP_ParseFaces();
	MAP_ParseBoundingBoxes();
	MAP_ParseLightmaps();
	MAP_ParseModels();

	MAP_ParseDummies();

	VERBOSE(1, ("map: parsed\n"));

	return 1;
}

static int MAP_Endianize(void)
{
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	unsigned char *tmp1;
	int i, j;
	unsigned long length, *tmp;
	bsp_texture_t *cur_texture;
	bsp_effect_t *cur_effect;
#endif

	/* Convert header */
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
	map.bsp.header->ident=SDL_SwapBE32(map.bsp.header->ident);
#else
	map.bsp.header->version=SDL_SwapLE32(map.bsp.header->version);
	for (i=0; i<FAKK_BSP_ENTRIES; i++) {
		map.bsp.header->entries[i].offset=SDL_SwapLE32(map.bsp.header->entries[i].offset);
		map.bsp.header->entries[i].length=SDL_SwapLE32(map.bsp.header->entries[i].length);
	}
#endif

	/* Convert textures */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_TEXTURES].length;
	cur_texture=(bsp_texture_t *)MAP_GetPointer(FAKK_TEXTURES);
	for (i=0; i<length/sizeof(bsp_texture_t); i++, cur_texture++) {
		cur_texture->flags=SDL_SwapLE32(cur_texture->flags);
		cur_texture->contents=SDL_SwapLE32(cur_texture->contents);
		cur_texture->dummy1=SDL_SwapLE32(cur_texture->dummy1);
	}
	VERBOSE(1, ("map: endianize: %lu textures done\n", length/sizeof(bsp_texture_t)));
#endif
	
	/* Convert planes */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_PLANES].length;
	tmp=(unsigned long *)MAP_GetPointer(FAKK_PLANES);
	for (i=0; i<length/sizeof(bsp_plane_t); i++) {
		for (j=0; j<sizeof(bsp_plane_t)/sizeof(unsigned long); j++,tmp++) {
			*tmp=SDL_SwapLE32(*tmp);
		}
	}
	VERBOSE(1, ("map: endianize: %lu planes done\n", length/sizeof(bsp_plane_t)));
#endif
	
	/* Convert lightmaps */
	/* R,G,B should be byteswapped on big-endian systems */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_LIGHTMAPS].length;
	tmp1=(unsigned char *)MAP_GetPointer(FAKK_LIGHTMAPS);
	for (i=0; i<length/3; i++, tmp+=3) {
		unsigned char color;

		color=tmp[0];
		tmp[0]=tmp[2];
		tmp[2]=color;
	}
	VERBOSE(1, ("map: endianize: lightmaps done\n"));
#endif

	/* Convert faces */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_FACES].length;
	tmp=(unsigned long *) MAP_GetPointer(FAKK_FACES);
	for (i=0; i<length/sizeof(bsp_face_t); i++) {
		for (j=0; j<sizeof(bsp_face_t)/sizeof(unsigned long); j++,tmp++) {
			*tmp=SDL_SwapLE32(*tmp);
		}
	}
	VERBOSE(1, ("map: endianize: %lu faces done\n", length/sizeof(bsp_face_t))); 
#endif
	
	/* Convert vertexes */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_VERTEXES].length;
	tmp=(unsigned long *) MAP_GetPointer(FAKK_VERTEXES);
	for (i=0; i<length/sizeof(bsp_vertex_t); i++) {
		for (j=0; j<sizeof(bsp_vertex_t)/sizeof(unsigned long); j++,tmp++) {
			*tmp=SDL_SwapLE32(*tmp);
		}
	}
	VERBOSE(1, ("map: endianize: %lu vertexes done\n", length/sizeof(bsp_vertex_t)));
#endif

	/* Convert meshverts */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_MESHVERTEXES].length;
	tmp=(unsigned long *) MAP_GetPointer(FAKK_MESHVERTEXES);
	for (i=0; i<length/sizeof(bsp_meshvertex_t); i++,tmp++) {
		*tmp=SDL_SwapLE32(*tmp);
	}
	VERBOSE(1, ("map: endianize: %lu meshvertexes done\n", length/sizeof(bsp_meshvertex_t))); 
#endif

	/* Convert leafbrushes */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_LEAFBRUSHES].length;
	tmp=(unsigned long *) MAP_GetPointer(FAKK_LEAFBRUSHES);
	for (i=0; i<length/sizeof(bsp_leafbrush_t); i++,tmp++) {
		*tmp=SDL_SwapLE32(*tmp);
	}
	VERBOSE(1, ("map: endianize: %lu leafbrushes done\n", length/sizeof(bsp_leafbrush_t))); 
#endif
	
	/* Convert leaffaces */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_LEAFFACES].length;
	tmp=(unsigned long *) MAP_GetPointer(FAKK_LEAFFACES);
	for (i=0; i<length/sizeof(bsp_leafface_t); i++,tmp++) {
		*tmp=SDL_SwapLE32(*tmp);
	}
	VERBOSE(1, ("map: endianize: %lu leaffaces done\n", length/sizeof(bsp_leafface_t))); 
#endif
	
	/* Convert leaves */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_LEAVES].length;
	tmp=(unsigned long *) MAP_GetPointer(FAKK_LEAVES);
	for (i=0; i<length/sizeof(bsp_leaf_t); i++) {
		for (j=0; j<sizeof(bsp_leaf_t)/sizeof(unsigned long); j++,tmp++) {
			*tmp=SDL_SwapLE32(*tmp);
		}
	}
	VERBOSE(1, ("map: endianize: %lu leaves done\n", length/sizeof(bsp_leaf_t))); 
#endif

	/* Convert nodes */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_NODES].length;
	tmp=(unsigned long *) MAP_GetPointer(FAKK_NODES);
	for (i=0; i<length/sizeof(bsp_node_t); i++) {
		for (j=0; j<sizeof(bsp_node_t)/sizeof(unsigned long); j++,tmp++) {
			*tmp=SDL_SwapLE32(*tmp);
		}
	}
	VERBOSE(1, ("map: endianize: %lu nodes done\n", length/sizeof(bsp_node_t))); 
#endif

	/* Convert brushsides */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_BRUSHSIDES].length;
	tmp=(unsigned long *) MAP_GetPointer(FAKK_BRUSHSIDES);
	for (i=0; i<length/sizeof(bsp_brushside_t); i++) {
		for (j=0; j<sizeof(bsp_brushside_t)/sizeof(unsigned long); j++,tmp++) {
			*tmp=SDL_SwapLE32(*tmp);
		}
	}
	VERBOSE(1, ("map: endianize: %lu brushsides done\n", length/sizeof(bsp_brushside_t))); 
#endif
	
	/* Convert brushes */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_BRUSHES].length;
	tmp=(unsigned long *) MAP_GetPointer(FAKK_BRUSHES);
	for (i=0; i<length/sizeof(bsp_brush_t); i++) {
		for (j=0; j<sizeof(bsp_brush_t)/sizeof(unsigned long); j++,tmp++) {
			*tmp=SDL_SwapLE32(*tmp);
		}
	}
	VERBOSE(1, ("map: endianize: %lu brushes done\n", length/sizeof(bsp_brush_t))); 
#endif
	
	/* Convert effects */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_EFFECTS].length;
	cur_effect=(bsp_effect_t *) MAP_GetPointer(FAKK_EFFECTS);
	for (i=0; i<length/sizeof(bsp_effect_t); i++, cur_texture++) {
		cur_effect->brush=SDL_SwapLE32(cur_effect->brush);
		cur_effect->dummy1=SDL_SwapLE32(cur_effect->dummy1);
	}
	VERBOSE(1, ("map: endianize: %lu effects done\n", length/sizeof(bsp_effect_t))); 
#endif
	
	/* Convert models */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	length=map.bsp.header->entries[FAKK_MODELS].length;
	tmp=(unsigned long *) MAP_GetPointer(FAKK_MODELS);
	for (i=0; i<length/sizeof(bsp_model_t); i++) {
		for (j=0; j<sizeof(bsp_model_t)/sizeof(unsigned long); j++,tmp++) {
			*tmp=SDL_SwapLE32(*tmp);
		}
	}
	VERBOSE(1, ("map: endianize: %lu models done\n", length/sizeof(bsp_model_t))); 
#endif	

	/* Convert entities */

	/* Convert visdata */
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
	tmp=(unsigned long *) MAP_GetPointer(FAKK_VISDATA);
	*tmp=SDL_SwapLE32(*tmp);tmp++;	/* n_vecs */
	*tmp=SDL_SwapLE32(*tmp);		/* sz_vecs */
	VERBOSE(1, ("map: endianize: visdata done\n")); 
#endif
	
	/* Convert lightvols */
	/* R,G,B should be byteswapped on big-endian systems */

	return 1;
}

static int MAP_ParseTextures(void)
{
#define MSG_TEX_LOADED	"map: texture '%s' loaded\n"

	unsigned long	length;
	bsp_texture_t	*cur_texture;
	int i;
	unsigned char	texname[TEXTURE_NAME_LEN];
	PHYSFS_sint64 tex_length;
	int num_tex_loaded;

	length = map.bsp.header->entries[FAKK_TEXTURES].length;

	map.num_textures = length/sizeof(bsp_texture_t);
	map.mem_textures=malloc(sizeof(void *)*map.num_textures);
	if (map.mem_textures==NULL) {
		fprintf(stderr, "map: textures: can not allocate memory for textures\n");
		return 0;
	}

	VERBOSE(1, ("map: loading textures\n"));

	cur_texture = map.bsp.textures;
	num_tex_loaded=0;
	for (i=0; i<map.num_textures; i++,cur_texture++) {
		unsigned char *p;
		unsigned char c;

#ifndef HAVE_SNPRINTF
		int namelen, maxlen;

		strncpy(texname, cur_texture->name, sizeof(texname));
		namelen=strlen(texname);
		maxlen=4;
		if (TEXTURE_NAME_LEN-namelen<maxlen) {
			maxlen=TEXTURE_NAME_LEN-namelen;
		}
		strncpy(&texname[namelen], ".ftx", maxlen);
#else
		snprintf(texname, sizeof(texname), "%s.ftx", cur_texture->name);
#endif
		texname[TEXTURE_NAME_LEN-1]='\0';

		map.mem_textures[i]=FS_Load(texname, &tex_length);
		if (map.mem_textures[i]!=NULL) {
			num_tex_loaded++;
			VERBOSE(2, (MSG_TEX_LOADED, texname));
			VERBOSE(3, ("map:  flags=0x%08lx, contents=0x%08lx\n",cur_texture->flags,cur_texture->contents));
			continue;
		}

		/* Convert filename to lowercase */
		p=texname;
		while (*p!='\0') {
			c=*p;				
			if ((c>='A') && (c<='Z')) {
				c=c-'A'+'a';
				*p=c;
			}
			p++;
		}

		map.mem_textures[i]=FS_Load(texname, &tex_length);
		if (map.mem_textures[i]!=NULL) {
			num_tex_loaded++;
			VERBOSE(2, (MSG_TEX_LOADED, texname));
			VERBOSE(3, ("map:  flags=0x%08lx, contents=0x%08lx\n",cur_texture->flags,cur_texture->contents));
		} else {
			VERBOSE(1, ("map: unable to load texture '%s'\n",texname));
			VERBOSE(3, ("map:  flags=0x%08lx, contents=0x%08lx\n",cur_texture->flags,cur_texture->contents));
		}
	}

	VERBOSE(1, ("map: %d/%d textures loaded\n", num_tex_loaded, map.num_textures));
	return 1;
}

static int MAP_ParseVertices(void)
{
	int i, num_vertexes;

	num_vertexes = map.bsp.header->entries[FAKK_VERTEXES].length/sizeof(bsp_vertex_t);

	for (i=0; i<num_vertexes; i++) {
		float x,y,z;
/*
		float r,g,b;
		unsigned long color;
*/
		x = map.bsp.vertexes[i].position.x;
		y = map.bsp.vertexes[i].position.y;
		z = map.bsp.vertexes[i].position.z;

		if (i==0) {
			map.minx = map.maxx = x;
			map.miny = map.maxy = y;
			map.minz = map.maxz = z;
		} else {
			if (map.minx>x)	map.minx=x;
			if (map.maxx<x) map.maxx=x;
			if (map.miny>y) map.miny=y;
			if (map.maxy<y) map.maxy=y;
			if (map.minz>z) map.minz=z;
			if (map.maxz<z) map.maxz=z;
		}

/*
		r = ((bsp_vertexes[i].color >> 24) & 255) * brightness;
		if (r>255) r=255;
		g = ((bsp_vertexes[i].color >> 16) & 255) * brightness;
		if (g>255) g=255;
		b = ((bsp_vertexes[i].color >> 8) & 255) * brightness;
		if (b>255) b=255;

		color = (bsp_vertexes[i].color & 255) |
			 (((unsigned char )r)<<24) |
			 (((unsigned char )g)<<16) |
			 (((unsigned char )b)<<8);
		bsp_vertexes[i].color = color;
*/
	}

	VERBOSE(1, ("map: %d vertices parsed\n", num_vertexes));
	VERBOSE(2, ("map:  minx=%.3f, maxx=%.3f\n",map.minx,map.maxx));
	VERBOSE(2, ("map:  miny=%.3f, maxy=%.3f\n",map.miny,map.maxy));
	VERBOSE(2, ("map:  minz=%.3f, maxz=%.3f\n",map.minz,map.maxz));

	return 1;
}

static int MAP_ParseEntities(void)
{
	unsigned long	length;
	unsigned char *p, c;
	int i, j, pos;
	entity_t *cur_entity;
	unsigned char entity_name[ENTITY_NAME_LEN], entity_value[ENTITY_NAME_LEN];

	length = map.bsp.header->entries[FAKK_ENTITIES].length;

	/* Firstly, count entities */
	map.num_entities=pos=0;
	p = map.bsp.entities;
	while (pos<=length) {
		c=*p++;
		pos++;
		if (c=='{') {
			map.num_entities++;
		}
	}

	map.mem_entities = (entity_t *)malloc(map.num_entities*sizeof(entity_t));
	if (map.mem_entities==NULL) {
		fprintf(stderr,"map: not enough memory for entities");
		return 0;
	}

	/* Now put entities in the array */
	VERBOSE(1, ("map: loading entities\n"));
	cur_entity = map.mem_entities;
	i=pos=0;
	p=map.bsp.entities;
	while (pos<length) {
		VERBOSE(2, ("map: parsing entity %d\n", i));

		/* Search start of entity: '{' */
		while ((*p++!='{') && (pos<length)) {
			pos++;
		}
		if (pos>=length) {
			break;
		}
		pos++;	/* skip '{' */
		VERBOSE(32, ("map: {\n"));

		/* Copy data in memory, till end of entity: '}' */
		memset(cur_entity, 0, sizeof(entity_t));

		while ((*p!='}') && (pos<length)) {

			memset(entity_name, 0, ENTITY_NAME_LEN);
			memset(entity_value, 0, ENTITY_NAME_LEN);

			/* Search starting name '"' */
			while ((*p!='"') && (*p!='}') && (pos<length)) {
				p++;
				pos++;
			}
			if ((pos>=length) || (*p=='}')) {
				break;
			}
			p++;
			pos++;	/* skip '"' */

			/* Parse name */
			j=0;
			while ((j<ENTITY_NAME_LEN) && (*p!='"') && (pos<length)) {
				entity_name[j++]=*p++;
				pos++;
			}			
			if (pos>=length) {
				break;
			}
			p++;
			pos++;	/* skip '"' */

			entity_name[ENTITY_NAME_LEN-1]='\0';

			/* Go to start of value */
			while ((*p++!='"') && (pos<length)) {
				pos++;
			}
			if (pos>=length) {
				break;
			}
			pos++;	/* skip '"' */

			/* Parse value */
			j=0;
			while ((j<ENTITY_NAME_LEN) && (*p!='"') && (pos<length)) {
				entity_value[j++]=*p++;
				pos++;
			}			
			if (pos>=length) {
				break;
			}
			p++;
			pos++;	/* skip '"' */

			entity_value[ENTITY_NAME_LEN-1]='\0';

			VERBOSE(3, ("map:  %s=%s\n",entity_name,entity_value));

			/* put the value at the right place */
			if (strcmp("classname",entity_name)==0) {
				strcpy(cur_entity->classname, entity_value);
			} else if (strcmp("model",entity_name)==0) {
				strcpy(cur_entity->model, entity_value);
			} else if (strcmp("targetname",entity_name)==0) {
				strcpy(cur_entity->targetname, entity_value);
			} else if (strcmp("target",entity_name)==0) {
				strcpy(cur_entity->target, entity_value);
			} else if (strcmp("thread",entity_name)==0) {
				strcpy(cur_entity->thread, entity_value);
			} else if (strcmp("killthread",entity_name)==0) {
				strcpy(cur_entity->killthread, entity_value);
			} else if (strcmp("pickup_thread",entity_name)==0) {
				strcpy(cur_entity->pickup_thread, entity_value);
			} else if (strcmp("noise",entity_name)==0) {
				strcpy(cur_entity->noise, entity_value);
			} else if (strcmp("modelname",entity_name)==0) {
				strcpy(cur_entity->modelname, entity_value);
			} else if (strcmp("anim",entity_name)==0) {
				strcpy(cur_entity->anim, entity_value);
			} else if (strcmp("spawntargetname",entity_name)==0) {
				strcpy(cur_entity->spawntargetname, entity_value);
			} else if (strcmp("map",entity_name)==0) {
				strcpy(cur_entity->map, entity_value);
			} else if (strcmp("sound_move",entity_name)==0) {
				strncpy(cur_entity->sound_move, entity_value, ENTITY_NAME_LEN);
				cur_entity->sound_move[ENTITY_NAME_LEN-1]='\0';
			} else if (strcmp("sound_stop",entity_name)==0) {
				strncpy(cur_entity->sound_stop, entity_value, ENTITY_NAME_LEN);
				cur_entity->sound_stop[ENTITY_NAME_LEN-1]='\0';
			} else if (strcmp("origin",entity_name)==0) {
				sscanf(entity_value, "%f %f %f", &(cur_entity->origin[0]), (&cur_entity->origin[1]), (&cur_entity->origin[2]));
			} else if (strcmp("_color",entity_name)==0) {
				sscanf(entity_value, "%f %f %f", (&cur_entity->color[0]), (&cur_entity->color[1]), (&cur_entity->color[2]));
			} else if (strcmp("speed",entity_name)==0) {
				sscanf(entity_value, "%f", &(cur_entity->speed));
			} else if (strcmp("scale",entity_name)==0) {
				sscanf(entity_value, "%f", &(cur_entity->scale));
			} else if (strcmp("light",entity_name)==0) {
				sscanf(entity_value, "%f", &(cur_entity->scale));
			} else if (strcmp("magnitude",entity_name)==0) {
				sscanf(entity_value, "%f", &(cur_entity->magnitude));
			} else if (strcmp("angle",entity_name)==0) {
				sscanf(entity_value, "%f", &(cur_entity->magnitude));
			} else if (strcmp("volume",entity_name)==0) {
				sscanf(entity_value, "%f", &(cur_entity->volume));
			} else if (strcmp("time",entity_name)==0) {
				sscanf(entity_value, "%f", &(cur_entity->time));
			} else if (strcmp("wait",entity_name)==0) {
				sscanf(entity_value, "%f", &(cur_entity->wait));
			} else if (strcmp("cap",entity_name)==0) {
				sscanf(entity_value, "%d", &(cur_entity->cap));
			} else if (strcmp("dmg",entity_name)==0) {
				sscanf(entity_value, "%d", &(cur_entity->dmg));
			} else if (strcmp("spawnflags",entity_name)==0) {
				sscanf(entity_value, "%d", &(cur_entity->spawnflags));
			} else if (strcmp("duration",entity_name)==0) {
				sscanf(entity_value, "%d", &(cur_entity->duration));
			} else if (strcmp("radius",entity_name)==0) {
				sscanf(entity_value, "%d", &(cur_entity->radius));
			} else if (strcmp("fov",entity_name)==0) {
				sscanf(entity_value, "%d", &(cur_entity->fov));
			} else {
				fprintf(stderr,"map:  unknown entity: %s=%s\n", entity_name, entity_value);
			}
		}
		if (pos>=length) {
			break;
		}
		VERBOSE(3, ("map: }\n"));

		cur_entity++;
		i++;
	}	
	VERBOSE(1, ("map: %d entities parsed\n", map.num_entities));
	return 1;
}

static int MAP_ParseFaces(void)
{
	unsigned long length;
	bsp_face_t *cur_face;
	int i, faces_type[FACE_LAST];

	length = map.bsp.header->entries[FAKK_FACES].length;
	map.num_faces=length/sizeof(bsp_face_t);

	cur_face=map.bsp.faces;
	faces_type[FACE_POLYGON]=
		faces_type[FACE_PATCH]=
		faces_type[FACE_MESH]=
		faces_type[FACE_BILLBOARD]=0;
	for (i=0; i<map.num_faces; i++, cur_face++) {
		faces_type[cur_face->type]++;
	}

	VERBOSE(1, ("map: %d faces parsed\n", map.num_faces));
	VERBOSE(2, ("map:  %d polygons, %d patches, %d meshes, %d billboards\n",
		faces_type[FACE_POLYGON], faces_type[FACE_PATCH],
		faces_type[FACE_MESH], faces_type[FACE_BILLBOARD]
	));
	return 1;
}

static int MAP_ParseBoundingBoxes(void)
{
	int num_nodes, num_leaves, num_bbox, i;
	bsp_node_t *cur_node;
	bsp_leaf_t *cur_leaf;
	bsp_model_t *cur_model;
	bbox_t *cur_bbox;

	num_nodes = map.bsp.header->entries[FAKK_NODES].length / sizeof(bsp_node_t); 
	num_leaves = map.bsp.header->entries[FAKK_LEAVES].length / sizeof(bsp_leaf_t);
	map.num_models = map.bsp.header->entries[FAKK_MODELS].length / sizeof(bsp_model_t);
	num_bbox = num_nodes + num_leaves + map.num_models;

	map.bsp_bbox = (bbox_t *)malloc(num_bbox*sizeof(bbox_t));
	if (map.bsp_bbox==NULL) {
		fprintf(stderr,"map: not enough memory for bounding boxes\n");
		return 0;
	}
	cur_bbox = map.bsp_bbox;

	/* nodes bounding boxes */
	map.bbox_nodes = cur_bbox;
	cur_node=map.bsp.nodes;
	for (i=0; i<num_nodes; i++, cur_node++, cur_bbox++) {
		cur_bbox->vertexes[0].x = cur_node->mins[0];
		cur_bbox->vertexes[0].y = cur_node->mins[1];
		cur_bbox->vertexes[0].z = cur_node->mins[2];

		cur_bbox->vertexes[1].x = cur_node->mins[0];
		cur_bbox->vertexes[1].y = cur_node->mins[1];
		cur_bbox->vertexes[1].z = cur_node->maxs[2];

		cur_bbox->vertexes[2].x = cur_node->maxs[0];
		cur_bbox->vertexes[2].y = cur_node->mins[1];
		cur_bbox->vertexes[2].z = cur_node->mins[2];

		cur_bbox->vertexes[3].x = cur_node->maxs[0];
		cur_bbox->vertexes[3].y = cur_node->mins[1];
		cur_bbox->vertexes[3].z = cur_node->maxs[2];

		cur_bbox->vertexes[4].x = cur_node->maxs[0];
		cur_bbox->vertexes[4].y = cur_node->maxs[1];
		cur_bbox->vertexes[4].z = cur_node->mins[2];

		cur_bbox->vertexes[5].x = cur_node->maxs[0];
		cur_bbox->vertexes[5].y = cur_node->maxs[1];
		cur_bbox->vertexes[5].z = cur_node->maxs[2];

		cur_bbox->vertexes[6].x = cur_node->mins[0];
		cur_bbox->vertexes[6].y = cur_node->maxs[1];
		cur_bbox->vertexes[6].z = cur_node->mins[2];

		cur_bbox->vertexes[7].x = cur_node->mins[0];
		cur_bbox->vertexes[7].y = cur_node->maxs[1];
		cur_bbox->vertexes[7].z = cur_node->maxs[2];
	}	

	/* leaves bounding boxes */
	map.bbox_leaves = cur_bbox;
	cur_leaf=map.bsp.leaves;
	for (i=0; i<num_leaves; i++, cur_leaf++, cur_bbox++) {
		cur_bbox->vertexes[0].x = cur_leaf->mins[0];
		cur_bbox->vertexes[0].y = cur_leaf->mins[1];
		cur_bbox->vertexes[0].z = cur_leaf->mins[2];

		cur_bbox->vertexes[1].x = cur_leaf->mins[0];
		cur_bbox->vertexes[1].y = cur_leaf->mins[1];
		cur_bbox->vertexes[1].z = cur_leaf->maxs[2];

		cur_bbox->vertexes[2].x = cur_leaf->maxs[0];
		cur_bbox->vertexes[2].y = cur_leaf->mins[1];
		cur_bbox->vertexes[2].z = cur_leaf->mins[2];

		cur_bbox->vertexes[3].x = cur_leaf->maxs[0];
		cur_bbox->vertexes[3].y = cur_leaf->mins[1];
		cur_bbox->vertexes[3].z = cur_leaf->maxs[2];

		cur_bbox->vertexes[4].x = cur_leaf->maxs[0];
		cur_bbox->vertexes[4].y = cur_leaf->maxs[1];
		cur_bbox->vertexes[4].z = cur_leaf->mins[2];

		cur_bbox->vertexes[5].x = cur_leaf->maxs[0];
		cur_bbox->vertexes[5].y = cur_leaf->maxs[1];
		cur_bbox->vertexes[5].z = cur_leaf->maxs[2];

		cur_bbox->vertexes[6].x = cur_leaf->mins[0];
		cur_bbox->vertexes[6].y = cur_leaf->maxs[1];
		cur_bbox->vertexes[6].z = cur_leaf->mins[2];

		cur_bbox->vertexes[7].x = cur_leaf->mins[0];
		cur_bbox->vertexes[7].y = cur_leaf->maxs[1];
		cur_bbox->vertexes[7].z = cur_leaf->maxs[2];
	}	

	/* models bounding boxes */
	map.bbox_models = cur_bbox;
	cur_model=map.bsp.models;
	for (i=0; i<map.num_models; i++, cur_model++, cur_bbox++) {
		cur_bbox->vertexes[0].x = cur_model->mins[0];
		cur_bbox->vertexes[0].y = cur_model->mins[1];
		cur_bbox->vertexes[0].z = cur_model->mins[2];

		cur_bbox->vertexes[1].x = cur_model->mins[0];
		cur_bbox->vertexes[1].y = cur_model->maxs[1];
		cur_bbox->vertexes[1].z = cur_model->mins[2];

		cur_bbox->vertexes[2].x = cur_model->maxs[0];
		cur_bbox->vertexes[2].y = cur_model->maxs[1];
		cur_bbox->vertexes[2].z = cur_model->mins[2];

		cur_bbox->vertexes[3].x = cur_model->maxs[0];
		cur_bbox->vertexes[3].y = cur_model->mins[1];
		cur_bbox->vertexes[3].z = cur_model->mins[2];

		cur_bbox->vertexes[4].x = cur_model->mins[0];
		cur_bbox->vertexes[4].y = cur_model->mins[1];
		cur_bbox->vertexes[4].z = cur_model->maxs[2];

		cur_bbox->vertexes[5].x = cur_model->mins[0];
		cur_bbox->vertexes[5].y = cur_model->maxs[1];
		cur_bbox->vertexes[5].z = cur_model->maxs[2];

		cur_bbox->vertexes[6].x = cur_model->maxs[0];
		cur_bbox->vertexes[6].y = cur_model->maxs[1];
		cur_bbox->vertexes[6].z = cur_model->maxs[2];

		cur_bbox->vertexes[7].x = cur_model->maxs[0];
		cur_bbox->vertexes[7].y = cur_model->mins[1];
		cur_bbox->vertexes[7].z = cur_model->maxs[2];
	}	

	VERBOSE(1, ("map: %d bounding boxes parsed\n", num_bbox));
	VERBOSE(2, ("map:  %d for nodes, %d for leaves, %d for models\n", num_nodes, num_leaves, map.num_models));
	return 1;
}

static int MAP_ParseLightmaps(void)
{
	unsigned long length;
	unsigned char *color;
	int i;

	length = map.bsp.header->entries[FAKK_LIGHTMAPS].length;

	map.num_lightmaps=length/sizeof(bsp_lightmap_t);

	/* Apply gamma correction to lightmaps */
	color=(unsigned char *)map.bsp.lightmaps;
	for (i=0; i<length/3; i++, color+=3) {
		float r,g,b;
		
		r = (float) color[0] * brightness;
		if (r>255) r=255;
		color[0]=(unsigned char) r;
		
		g = (float) color[1] * brightness;
		if (g>255) g=255;
		color[1]=(unsigned char) g;
		
		b = (float) color[2] * brightness;
		if (b>255) b=255;
		color[2]=(unsigned char) b;
	}

	VERBOSE(1, ("map: %d lightmaps parsed\n", map.num_lightmaps));
	return 1;
}

static int MAP_ParseModels(void)
{
	entity_t	*cur_entity;
	unsigned long length;
	int i, num_model;

	length = map.bsp.header->entries[FAKK_MODELS].length;
	map.num_models=length/sizeof(bsp_model_t);

	map.model_to_entity=(entity_t **)malloc(sizeof(entity_t **)*map.num_models);
	if (map.model_to_entity==NULL) {
		fprintf(stderr,"map: not enough memory for model_to_entity\n");
		return 0;
	}

	cur_entity=map.mem_entities;
	for (i=0; i<map.num_entities; i++,cur_entity++) {
		if (cur_entity->model[0]!='*') {
			continue;
		}

		sscanf(&(cur_entity->model[1]), "%d", &num_model);

		if (num_model==0) {
			continue;
		}

		map.model_to_entity[num_model]=cur_entity;
	}

	VERBOSE(1, ("map: %d models parsed\n", map.num_models));
	return 1;
}

static int MAP_ParseDummies(void)
{
/*
	long	*cur_dummy;
	unsigned long length, num_dummies;
	int i,j, num_longs;
	long mins[15],maxs[15];

	length = map.bsp.header->entries[FAKK_BSP_17].length;
	num_dummies=length/sizeof(bsp_bsp17_t);
	num_longs=sizeof(bsp_bsp17_t)/sizeof(long);

	cur_dummy = (long *)MAP_GetPointer(FAKK_BSP_17);
	for (i=0; i<num_dummies; i++) {
		if (i==0) {
			for (j=0; j<num_longs; j++) {
				mins[j]=maxs[j]=cur_dummy[j];
			}
		} else {
			for (j=0; j<num_longs; j++) {
				if (cur_dummy[j]<mins[j]) mins[j]=cur_dummy[j];
				if (cur_dummy[j]>maxs[j]) maxs[j]=cur_dummy[j];
			}
		}

		cur_dummy+=num_longs;
	}

	VERBOSE(1, ("map: dummies parsed\n"));

	for (j=0; j<num_longs; j++) {
		float *f1,*f2;
		f1=&(mins[j]);
		f2=&(maxs[j]);
		VERBOSE(2, ("map:  %d: min=%i (%.3f), max=%i (%.3f)\n",j,mins[j],*f1,maxs[j],*f2));
	}
*/
	return 1;
}

float MAP_GetMaxDist(void)
{
	return CALC_GetLength(map.maxx-map.minx, map.maxy-map.miny, map.maxz-map.minz);
}

