/*
	Viewer

	Copyright (C) 2003	Patrice Mandin
	
	See COPYING in top directory for license
*/

/*--- Includes ---*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <math.h>
#include <stdlib.h> 
#include <string.h>
#include <unistd.h>

#include <SDL/SDL.h>
#include <SDL/SDL_opengl.h>

#include "calculations_types.h"
#include "camera.h"
#include "extensions.h"
#include "triedre.h"
#include "map_parse.h"
#include "map_draw.h"
#include "parameters.h"
#include "statistics.h"
#include "verbose.h"
#include "viewer.h"

/*--- Defines ---*/

#define VIEWER_FOV		60.0
#define FRAME_PER_TIC	10

#define MOVEMENT_BASE 	4.0
#define MOVEMENT_SCALE	4.0
#define TARGET_SCALE 200.0

#define ZNEAR 4.0

/*--- Global variables ---*/

int polyfill_type;
int model_selected;
int patch_selected;
SDL_bool models_enabled;

/*--- Variables ---*/

/* Screen properties */
static Uint32 videoflags;

/* Mouse */
static int Mousex, Mousey, Mouseb[3];
static float move_scale;

/* View properties */
static SDL_bool fog_enabled;
static SDL_bool zbuf_enabled;
static SDL_bool fps_enabled;
static SDL_bool tex_enabled;
static SDL_bool grab_enabled;
static SDL_bool dither_enabled;

/* Camera orientation */
static float anglex,angley,anglez;

/*--- Functions prototypes ---*/

static void VIEW_Init(void);
static void VIEW_Shutdown(void);
static int VIEW_ScanKeyboard(void);

static void VIEW_Movement(void);
static void VIEW_Draw(void);

/*--- Functions ---*/

int VIEW_Demo(void)
{
	int prevticks, frame_per_sec, frame_tick;
	int quit_demo;

	/* Initialize SDL for video output */
	if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
		fprintf(stderr, "Unable to initialize SDL: %s\n", SDL_GetError());
		return 0;
	}
	atexit(SDL_Quit);

	/* Create an OpenGL screen */
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
/*	SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE,32);*/

	videoflags = SDL_OPENGL;
	if (start_fullscreen) {
		videoflags |= SDL_FULLSCREEN;
	}
	if ( SDL_SetVideoMode(screen_width, screen_height, 0, videoflags) == NULL) {
		/* Seems win32 does not appreciate passing 0 as bpp */
		if ( SDL_SetVideoMode(screen_width, screen_height, 16, videoflags) == NULL) {
			fprintf(stderr, "Unable to create OpenGL screen: %s\n", SDL_GetError());
			return 0;
		}
	}

	SDL_WM_SetCaption("American McGee's Alice viewer "VERSION, NULL); 

	if ((videoflags & SDL_FULLSCREEN)==SDL_FULLSCREEN) {
		SDL_WM_GrabInput(SDL_GRAB_ON);
		SDL_ShowCursor(SDL_DISABLE);
		grab_enabled=SDL_TRUE;
	} else {
		grab_enabled=SDL_FALSE;
	}

	glViewport(0, 0, screen_width, screen_height);
	EXT_Check();
	VIEW_Init();
	TRI_Build();
	MAP_InitDraw();

	/* Init camera */
	MAP_InitCamera(&camera, &anglex);
	angley=anglez=0;
	CAM_CalcTarget(TARGET_SCALE, anglex, angley);
	glMatrixMode(GL_MODELVIEW);
	{
		glLoadIdentity();
		gluLookAt(
			camera.from.x, camera.from.y, camera.from.z,
			camera.to.x, camera.to.y, camera.to.z,
			0.0, 0.0, 1.0
		);
	}
	CAM_UpdateFrustum();

	/* Init mouse movement */
	Mousex = Mousey = Mouseb[0] = Mouseb[1] = Mouseb[2] = 0;
	move_scale=1.0;
	prevticks = SDL_GetTicks()/FRAME_PER_TIC;

	frame_per_sec = 0;
	frame_tick = SDL_GetTicks();

	quit_demo=0;
	while (!quit_demo) {
		int i, ticks;
		int curticks = SDL_GetTicks()/FRAME_PER_TIC;

		if (grab_enabled) {
			for (i=prevticks;i<curticks;i++) {
				SDL_GetRelativeMouseState(&Mousex, &Mousey);
				VIEW_Movement();
			}
		}
		prevticks=SDL_GetTicks()/FRAME_PER_TIC;

		VIEW_Draw();

		ticks = SDL_GetTicks();
		if (ticks-frame_tick > 1000) {
			frame_tick=ticks;
			if (fps_enabled) {
				printf("view: %d fps\n",frame_per_sec);
				STAT_Print();
			}
			frame_per_sec=0;
		}

		SDL_GL_SwapBuffers();
#ifndef WIN32
		usleep(1000);
#endif
		frame_per_sec++;

		quit_demo=VIEW_ScanKeyboard();
	}
	
	TRI_Destroy();
	VIEW_Shutdown();

	SDL_Quit();

	return 1;
}

static void VIEW_Init(void)
{
	float maxdist, mindist;

	glClearColor(0.0, 0.0, 0.0, 0.0);

	/* Projection matrix */
	mindist = ZNEAR;
	maxdist = MAP_GetMaxDist();

	glMatrixMode(GL_PROJECTION);
	{
		glLoadIdentity();
		gluPerspective(VIEWER_FOV, (GLfloat) screen_width / (GLfloat) screen_height,  mindist, maxdist);
	}

	/* Texturing, blending */
	if (arb_multitexture_present) {
		if (multitexture_units>=2) {
			glActiveTextureARB(GL_TEXTURE1_ARB);
			glMatrixMode(GL_TEXTURE);
			{
				glLoadIdentity();
			}
			glDisable(GL_TEXTURE_2D);
			glDisable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		}
		glActiveTextureARB(GL_TEXTURE0_ARB);
	}
	glMatrixMode(GL_TEXTURE);
	{
		glLoadIdentity();
	}
	glDisable(GL_TEXTURE_2D);
	glDisable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	/* Dithering */
	glEnable(GL_DITHER);

	/* Z-Buffer */
	glEnable(GL_DEPTH_TEST);
	{
		glDepthFunc(GL_LEQUAL);
	}

	/* Face culling */
	glEnable(GL_CULL_FACE);
	{
		glFrontFace(GL_CW);
	}

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_COLOR_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
	if (arb_multitexture_present) {
		if (multitexture_units>=2) {
			glClientActiveTextureARB(GL_TEXTURE1_ARB);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		}
		glClientActiveTextureARB(GL_TEXTURE0_ARB);
	}
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);

	fog_enabled = SDL_FALSE;
	zbuf_enabled = SDL_TRUE;
	fps_enabled = SDL_FALSE;
	tex_enabled = SDL_TRUE;
	dither_enabled = SDL_TRUE;

	polyfill_type = POLYFILL_TEXMODVER;
	models_enabled = SDL_TRUE;
	model_selected= -1 /*48*/;
	patch_selected= -1 /*279*/;

	{
		int eval_max_order;
		
		glGetIntegerv(GL_MAX_EVAL_ORDER, &eval_max_order);
		VERBOSE(1, ("gl: maximum evaluator order: %d\n", eval_max_order));
	}
}

static void VIEW_Shutdown(void)
{
	glDisableClientState(GL_COLOR_ARRAY);
	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);
	if (arb_multitexture_present) {
		if (multitexture_units>=2) {
			glClientActiveTextureARB(GL_TEXTURE1_ARB);
			glDisableClientState(GL_TEXTURE_COORD_ARRAY);
		}
		glClientActiveTextureARB(GL_TEXTURE0_ARB);
	}
	glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	glDisable(GL_CULL_FACE);
	glDisable(GL_DEPTH_TEST);
}

static int VIEW_ScanKeyboard(void)
{
	SDL_Event event;

	while ( SDL_PollEvent(&event) ) {
		switch(event.type) {
			case SDL_QUIT:
				return 1;
				break;
			case SDL_KEYDOWN:
				switch (event.key.keysym.sym) {
					/* Escape to quit demo */
					case SDLK_ESCAPE:
						return 1;
						break;
					/* ALT-Enter: toggle fullscreen */
					case SDLK_RETURN:
						if (event.key.keysym.mod & KMOD_ALT) {
							videoflags ^= SDL_FULLSCREEN;
							SDL_SetVideoMode(screen_width, screen_height, 0, videoflags);
							if ((videoflags & SDL_FULLSCREEN)==SDL_FULLSCREEN) {
								SDL_WM_GrabInput(SDL_GRAB_ON);
								SDL_ShowCursor(SDL_DISABLE);
								grab_enabled=SDL_TRUE;
							}
						}
						break;
					/* Shift ? */
					case SDLK_RSHIFT:
					case SDLK_LSHIFT:
						move_scale=MOVEMENT_SCALE;
						break;
					/* C to display coordinates */
					case SDLK_c:
						{
							printf("view: X=%5.3f, Y=%5.3f, Z=%5.3f, AX=%5.3f, AY=%.3f\n",camera.from.x,camera.from.y,camera.from.z,anglex,angley);
						}
						break;
					/* D to toggle dithering */
					case SDLK_d:
						dither_enabled ^= 1;
						if (dither_enabled) {
							VERBOSE(1, ("view: dithering enabled\n"));
						} else {
							VERBOSE(1, ("view: dithering disabled\n"));
						}
						break;
					/* G to toggle mousegrab */
					case SDLK_g:
						if ((videoflags & SDL_FULLSCREEN)!=SDL_FULLSCREEN) {
							grab_enabled ^= 1;
							if (grab_enabled) {
								VERBOSE(1, ("view: grab enabled\n"));
								SDL_WM_GrabInput(SDL_GRAB_ON);
								SDL_ShowCursor(SDL_DISABLE);
							} else {
								VERBOSE(1, ("view: grab disabled\n"));
								SDL_WM_GrabInput(SDL_GRAB_OFF);
								SDL_ShowCursor(SDL_ENABLE);
							}
						}
						break;
					/* M to toggle models display */
					case SDLK_m:
						models_enabled ^= 1;
						break;
					/* S to toggle fps display */
					case SDLK_s:
						fps_enabled ^= 1;
						break;
					/* Z to toggle depth buffer */
					case SDLK_z:
						zbuf_enabled ^= 1;
						if (zbuf_enabled) {
							VERBOSE(1, ("view: zbuffer enabled\n"));
						} else {
							VERBOSE(1, ("view: zbuffer disabled\n"));
						}
						break;
					/* F1-F2 to change polygon filling type */
					case SDLK_F1:
						--polyfill_type;
						if (polyfill_type<0) polyfill_type=0;
						switch (polyfill_type) {
							case POLYFILL_VERTEXCOLOR:
								VERBOSE(1, ("view: polygon filling: vertices colors\n"));
								break;
							case POLYFILL_TEXTURE:
								VERBOSE(1, ("view: polygon filling: texture\n"));
								break;
							case POLYFILL_TEXMODVER:
								VERBOSE(1, ("view: polygon filling: texture modulated with vertices colors\n"));
								break;
							case POLYFILL_LIGHTMAP:
								VERBOSE(1, ("view: polygon filling: lightmap\n"));
								break;
							case POLYFILL_TEXMODLIGHT:
								VERBOSE(1, ("view: polygon filling: texture modulated with lightmap\n"));
								break;
							default:
								break;
						}
						MAP_RefreshDraw();
						break;
					case SDLK_F2:
						++polyfill_type;
						if (polyfill_type>=POLYFILL_LAST) polyfill_type=POLYFILL_LAST-1;
						switch (polyfill_type) {
							case POLYFILL_VERTEXCOLOR:
								VERBOSE(1, ("view: polygon filling: vertices colors\n"));
								break;
							case POLYFILL_TEXTURE:
								VERBOSE(1, ("view: polygon filling: texture\n"));
								break;
							case POLYFILL_TEXMODVER:
								VERBOSE(1, ("view: polygon filling: texture modulated with vertices colors\n"));
								break;
							case POLYFILL_LIGHTMAP:
								VERBOSE(1, ("view: polygon filling: lightmap\n"));
								break;
							case POLYFILL_TEXMODLIGHT:
								VERBOSE(1, ("view: polygon filling: texture modulated with lightmap\n"));
								break;
							default:
								break;
						}
						MAP_RefreshDraw();
						break;
					case SDLK_F3:
						--model_selected;
						if (model_selected<-1) model_selected=-1;
						VERBOSE(1, ("view: model selected: %d\n", model_selected));
						break;
					case SDLK_F4:
						++model_selected;
						if (model_selected>=map.num_models) model_selected=map.num_models-1;
						VERBOSE(1, ("view: model selected: %d\n", model_selected));
						break;
					case SDLK_F5:
						--patch_selected;
						if (patch_selected<-1) patch_selected=-1;
						VERBOSE(1, ("view: patch selected: %d\n", patch_selected));
						break;
					case SDLK_F6:
						++patch_selected;
						if (patch_selected>=map.num_patches) patch_selected=map.num_patches-1;
						VERBOSE(1, ("view: patch selected: %d\n", patch_selected));
						break;
					default:
						break;
				}
				break;
			case SDL_KEYUP:
				switch (event.key.keysym.sym) {
					/* Shift ? */
					case SDLK_RSHIFT:
					case SDLK_LSHIFT:
						move_scale=1.0;
						break;
					default:
						break;
				}
				break;
			case SDL_MOUSEBUTTONDOWN:
				Mouseb[0]=(SDL_BUTTON_LEFT == event.button.button);
				Mouseb[1]=(SDL_BUTTON_MIDDLE == event.button.button);
				Mouseb[2]=(SDL_BUTTON_RIGHT == event.button.button);
				break;
			case SDL_MOUSEBUTTONUP:
				Mouseb[0] = Mouseb[1] = Mouseb[2] = 0;
				break;
		}
	}

	return 0;
}

static void VIEW_Movement(void)
{
	int do_movement;
	float vax,vay;

	do_movement=0;

	/* Left button to move forward in space */
	if (Mouseb[0]) {
		vax=(anglex * M_PI) / 180.0;
		vay=(angley * M_PI) / 180.0;

		camera.from.x+=MOVEMENT_BASE*move_scale*cos(vax)*cos(-vay);
		camera.from.y+=MOVEMENT_BASE*move_scale*-sin(vax)*cos(-vay);
		camera.from.z+=MOVEMENT_BASE*move_scale*sin(-vay);

		do_movement=1;
	}

	/* Right button, reset values */
	if (Mouseb[2]) {
		MAP_InitCamera(&camera,&anglex);
		angley=anglez=0;

		do_movement=1;
	}

	/* Mouse movement */
	if (Mousex) {
		anglex += Mousex/2.0;
		do_movement=1;
	}	
	
	if (Mousey) {
		angley += Mousey/2.0;
		if (angley<-85.0) angley=-85.0;
		if (angley>85.0) angley=85.0;
		do_movement=1;
	}

	if (do_movement) {
		CAM_CalcTarget(TARGET_SCALE, anglex, angley);

		glMatrixMode(GL_MODELVIEW);
		{
			glLoadIdentity();
			gluLookAt(
				camera.from.x, camera.from.y, camera.from.z,
				camera.to.x, camera.to.y, camera.to.z,
				0.0, 0.0, 1.0
			);
		}

		CAM_UpdateFrustum();
	}
}

static void VIEW_Draw(void)
{
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

	if (zbuf_enabled) {
		glEnable(GL_DEPTH_TEST);
	} else {
		glDisable(GL_DEPTH_TEST);
	}

	if (dither_enabled) {
		glEnable(GL_DITHER);
	} else {
		glDisable(GL_DITHER);
	}

	/*--- Draw triedre ---*/
#if 0
	glDisable(GL_TEXTURE_2D);
	TRI_Display();
#endif

	/*--- Draw map ---*/
	MAP_Draw();
}
